from viper import Host, Hosts
from viper.demo import __doc__

group1 = Hosts.from_items(
    Host("127.0.0.1"),
    Host("1.1.1.1"),
    Host("2.2.2.2"),
    Host("8.8.8.8"),
    Host(
        ip="::1",
        hostname="localhost",
        domain="localdomain",
        port=22,
        login_name="root",
        identity_file="/root/.ssh/id_rsa.pub",
    ),
)



import typing as t

from viper import Host, Task
from viper.demo import __doc__


def ping_command(host: Host) -> t.Sequence[str]:
    return ("ping", "-c", "1", host.ip)


ping = Task("Ping", ping_command, timeout=5, retry=1)



from viper import TaskResults
from viper.demo import __doc__


def print_status(results: TaskResults) -> str:
    """Print the status of the results in the terminal."""

    for result in results.all():
        if result.errored():
            print(f"{result.task.name}: {result.host.ip} FAILED")
        else:
            print(f"{result.task.name}: {result.host.ip} PASSED")



from __future__ import annotations

import sqlite3
import typing as t
from dataclasses import dataclass, field

from viper.const import Config


@dataclass
class ViperDB:

    url: str = Config.db_url.value
    engine: t.Optional[sqlite3.Connection] = field(init=False, default=None)

    @classmethod
    def init(cls, url, force=False):
        if force:
            with cls(url) as conn:
                conn.execute("DELETE FROM task_results")
                conn.execute("DROP TABLE task_results")

        with cls(url) as conn:
            conn.execute(
                """
                CREATE TABLE IF NOT EXISTS task_results (
                    id INTEGER PRIMARY KEY AUTOINCREMENT,
                    hash INTEGER UNIQUE NOT NULL,
                    task JSON NOT NULL,
                    host JSON NOT NULL,
                    command JSON NOT NULL,
                    stdout BLOB NOT NULL,
                    stderr BLOB NOT NULL,
                    returncode INTEGER NOT NULL,
                    start REAL NOT NULL,
                    end REAL NOT NULL,
                    retry INTEGER NOT NULL
                );
                """
            )

    def __enter__(self):
        self.engine = sqlite3.connect(self.url)
        return self.engine.cursor()

    def __exit__(self, exc_type, exc_value, exc_traceback):
        try:
            self.engine.commit()
        except Exception:
            self.engine.rollback()
        self.engine.close()



__author__ = "Arijit Basu"
__description__ = "Viper is a handy tool for easily running infrastructure management tasks and commands."
__email__ = "sayanarijit@gmail.com"
__homepage__ = "https://github.com/sayanarijit/viper"
__license__ = "MIT"
__version__ = "v0.0.1"

from viper.collections import (
    Host,
    Hosts,
    Task,
    TaskResult,
    TaskResults,
    TaskRunner,
    TaskRunners,
)
from viper.demo import __doc__ as __viper_cli_examples__

__doc__ = f"""{__description__}

{__viper_cli_examples__}
"""



"""Base collection classes are defined here."""

from __future__ import annotations

import time
import typing as t
from concurrent.futures import ThreadPoolExecutor, as_completed
from dataclasses import dataclass, field
from json import dumps as dumpjson
from json import loads as loadjson
from pydoc import locate
from subprocess import PIPE, Popen, TimeoutExpired

from viper.const import Config
from viper.db import ViperDB


@dataclass(frozen=True, order=True)
class Item:
    """A single item."""

    @classmethod
    def from_dict(cls, dict_: t.Dict[str, t.Any]) -> Item:
        """Initialize item from given dict."""

        try:
            return cls(**dict_)
        except Exception:
            raise ValueError(f"invalid input data for {cls.__name__}")

    def to_dict(self) -> t.Dict[str, t.Any]:
        """Represent the item as dict."""

        return vars(self)

    @classmethod
    def from_json(cls, json: str, *args: t.Any, **kwargs: t.Any) -> Item:
        """Initialize item from given JSON data."""

        return cls.from_dict(loadjson(json))

    def to_json(self, *args: t.Any, **kwargs: t.Any) -> str:
        """Represent the item as JSON data."""

        return dumpjson(self.to_dict(), *args, **kwargs)

    @classmethod
    def from_obj(cls, objpath: str) -> Item:
        """Load the item from the given Python object path."""

        item = locate(objpath)
        if not item:
            raise ValueError(f"could not resolve {repr(objpath)}.")

        if not isinstance(item, cls):
            raise ValueError(f"{repr(objpath)} is not a valid {cls} instance.")

        return item

    def hash(self):
        """Get the hash value"""
        return hash(self)

    def pipe(self, func: t.Callable[[Item], t.Any]) -> t.Any:
        """Pipe this object to the given function"""
        return func(self)


@dataclass(frozen=True)
class Items:
    """A collection of similar items."""

    _all: t.Sequence[Item] = ()
    _item_factory: t.Type[Item] = field(init=False, default=Item)

    @classmethod
    def from_items(cls, *items: Item) -> Items:
        """Initialize items from given items."""
        return cls(tuple(set(items)))

    @classmethod
    def from_list(cls, list_: t.Sequence[t.Dict[str, t.Any]]) -> Items:
        """Initialize items from given list."""

        if cls._item_factory is None:
            raise NotImplementedError()

        try:
            return cls.from_items(*map(cls._item_factory.from_dict, list_))
        except Exception:
            raise ValueError(f"invalid input data for {cls.__name__}")

    def to_list(self) -> t.List[t.Dict[str, t.Any]]:
        """Represent the items as list."""

        return [i.to_dict() for i in self._all]

    @classmethod
    def from_json(cls, json: str, *args: t.Any, **kwargs: t.Any) -> Items:
        """Initialize items from given JSON data."""

        return cls.from_list(loadjson(json))

    def to_json(self, *args: t.Any, **kwargs: t.Any) -> str:
        """Represent the item as JSON data."""

        return dumpjson(self.to_list(), *args, **kwargs)

    @classmethod
    def from_obj(cls, objpath: str) -> Items:
        """Load the items from the given Python object path."""

        items = locate(objpath)
        if not items:
            raise ValueError(f"could not resolve {repr(objpath)}.")

        if not isinstance(items, cls):
            raise ValueError(f"{repr(objpath)} is not a valid {cls} instance.")

        return items

    def __getitem__(self, key: int) -> Item:
        return self._all[key]

    def __len__(self) -> int:
        """Count the number of items."""
        return len(self._all)

    def count(self) -> int:
        """Count the number of items."""
        return len(self._all)

    def first(self) -> Item:
        """Get the first item from the list."""
        return self._all[0]

    def last(self) -> Item:
        """Get the last item from the list."""
        return self._all[-1]

    def index(self, index: int) -> Item:
        """The the item from a given index."""
        return self._all[index]

    def sort(self, key: t.Optional[t.Callable[[Item], t.Any]] = None) -> Items:
        """Sort the items by given key/function."""
        return type(self)(tuple(sorted(self._all, key=key)))

    def filter(self, func: t.Callable[[Item], bool]) -> Items:
        """Filter the items by a giver function."""
        return type(self)(tuple(filter(func, self._all)))

    def get(self, func: t.Callable[[Item], bool]) -> Item:
        """Get one unique item by a filter function."""

        items = self.filter(func)

        if items.count() == 0:
            raise LookupError("could not find any item.")

        if items.count() > 1:
            raise LookupError("multiple item found.")

        return items.first()

    def all(self) -> t.Sequence[Item]:
        """Get a tuple of all the items."""
        return self._all

    def hash(self):
        """Get the hash value"""
        return hash(self)

    def pipe(self, func: t.Callable[[Items], t.Any]) -> t.Any:
        """Pipe this object to the given function"""
        return func(self)


@dataclass(frozen=True, order=True)
class Host(Item):
    """Viper Host class."""

    ip: str
    hostname: t.Optional[str] = None
    domain: t.Optional[str] = None
    port: int = 22
    login_name: t.Optional[str] = None
    identity_file: t.Optional[str] = None

    def fqdn(self) -> str:
        """Get the FQDN from hostname and domainname."""

        if not self.hostname and not self.domain:
            raise AttributeError("hostname and domain not set")

        if not self.hostname:
            raise AttributeError("hostname not set")

        if not self.domain:
            raise AttributeError("domain not set")

        return f"{self.hostname}.{self.domain}"

    def task(self, task: Task) -> TaskRunner:
        """Assigns a task to be run."""

        return TaskRunner(task=task, host=self)

    def run_task(self, task: Task) -> TaskResult:
        """Assign the task to the host and then run it."""

        return self.task(task).run()

    def task_results(self) -> TaskResults:
        """Fetch recent task results of current host from database."""
        return self.pipe(TaskResults.by_host)


@dataclass(frozen=True)
class Hosts(Items):
    """Hosts manager."""

    _item_factory: t.Type[Host] = field(init=False, default=Host)

    @classmethod
    def from_file(
        cls, filepath: str, loader: t.Callable[[t.TextIO], "Hosts"] = None
    ) -> Hosts:
        """Initialize hosts reading from a file."""

        if loader is None:

            def _loader(f: t.TextIO):
                return cls.from_items(
                    *(Host(ip.strip()) for ip in f.read().strip().split())
                )

            loader = _loader

        with open(filepath) as f:
            return loader(f)

    def task(self, task: Task) -> TaskRunners:
        """Assigns a task to be run on all the hosts."""

        return TaskRunners.from_items(
            *(TaskRunner(task=task, host=h) for h in self._all)
        )

    def run_task(self, task: Task, max_workers=Config.max_workers.value) -> TaskResults:
        """Run a task to be run on all hosts and then run it."""

        return self.task(task).run(max_workers=max_workers)

    def run_task_then_pipe(
        self,
        task: Task,
        func: t.Callable[[TaskResults], t.Any],
        max_workers=Config.max_workers.value,
    ) -> t.Any:
        """Assign the task to the host and then run it."""

        return self.run_task(task, max_workers=max_workers).pipe(func)

    def task_results(self) -> TaskResults:
        results = []
        for h in self._all:
            for r in h.task_results():
                results.append(r)
        return TaskResults.from_items(*results)


@dataclass(frozen=True, order=True)
class Task(Item):
    """An infra task."""

    name: str
    command_factory: t.Callable[[Host], t.Sequence[str]]
    timeout: t.Optional[int] = None
    retry: int = 0
    stdout_processor: t.Optional[t.Callable[[str], str]] = None
    stderr_processor: t.Optional[t.Callable[[str], str]] = None

    @classmethod
    def from_dict(cls, dict_: t.Dict[str, t.Any]) -> Task:
        """Overloading from_dict()."""

        outp = dict_.get("stdout_processor")
        errp = dict_.get("stderr_processor")

        return cls(
            **dict(
                dict_,
                command_factory=locate(dict_["command_factory"]),
                stdout_processor=locate(outp) if outp else None,
                stderr_processor=locate(errp) if errp else None,
            )
        )

    def to_dict(self) -> t.Dict[str, t.Any]:
        """Overloading to_dict()."""

        cf = self.command_factory
        outp = self.stdout_processor
        errp = self.stderr_processor

        return dict(
            vars(self),
            command_factory=f"{cf.__module__}.{cf.__qualname__}",
            stdout_processor=f"{outp.__module__}.{outp.__qualname__}" if outp else None,
            stderr_processor=f"{errp.__module__}.{errp.__qualname__}" if errp else None,
        )

    def results(self) -> TaskResults:
        """Get the past results of this task."""

        return self.pipe(TaskResults.by_task)


@dataclass(frozen=True, order=True)
class TaskRunner(Item):
    """A task runner."""

    host: Host
    task: Task

    @classmethod
    def from_dict(cls, dict_: t.Dict[str, t.Any]) -> TaskRunner:
        """Overloading from_dict()."""

        return cls(
            **dict(
                dict_,
                task=Task.from_dict(dict_["task"]),
                host=Host.from_dict(dict_["host"]),
            )
        )

    def to_dict(self) -> t.Dict[str, t.Any]:
        """Overloading to_dict()."""

        return dict(vars(self), task=self.task.to_dict(), host=self.host.to_dict())

    def run(self, retry=0) -> TaskResult:
        """Run the task on the host."""
        command = self.task.command_factory(self.host)

        p = Popen(command, stdout=PIPE, stderr=PIPE)

        start = time.time()
        try:
            out, err = p.communicate(timeout=self.task.timeout)
            stdout, stderr = out.decode("latin1"), err.decode("latin1")
        except TimeoutExpired as e:
            stdout, stderr, p.returncode = "", str(e), 123
        end = time.time()

        if self.task.stderr_processor:
            stdout = self.task.stdout_processor(stdout)

        if self.task.stderr_processor:
            stderr = self.task.stderr_processor(stderr)

        result = TaskResult(
            self.task,
            self.host,
            command,
            stdout,
            stderr,
            p.returncode,
            start,
            end,
            retry,
        ).save()

        if result.errored() and self.task.retry > retry:
            return self.run(retry=retry + 1)

        return result

    def task_results(self) -> TaskResults:
        """Fetch recent task results of current task runner from database."""
        return self.pipe(TaskResults.by_task_runner)


@dataclass(frozen=True)
class TaskRunners(Items):
    _item_factory: t.Type[TaskRunner] = field(init=False, default=TaskRunner)

    def run(self, max_workers=Config.max_workers.value) -> TaskResults:
        """Run the tasks."""

        if max_workers <= 1:
            # Run in sequence
            return TaskResults.from_items(*(r.run() for r in self._all))

        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            # Run in parallel
            futures = [executor.submit(r.run) for r in self._all]
            results = [f.result() for f in as_completed(futures)]

        return TaskResults.from_items(*results)

    def hosts(self) -> Hosts:
        """Get the list of hosts from the runners."""
        return Hosts.from_items(*(t.host for t in self._all))


@dataclass(frozen=True, order=True)
class TaskResult(Item):
    """The result of an executed task."""

    task: Task
    host: Host
    command: t.Sequence[str]
    stdout: str
    stderr: str
    returncode: int
    start: float
    end: float
    retry: int

    @classmethod
    def from_hash(cls, hash_: int) -> TaskResult:
        with ViperDB(ViperDB.url) as conn:

            data = next(
                conn.execute(
                    """
                    SELECT
                        task, host, command, stdout, stderr, returncode, start, end, retry
                    FROM task_results WHERE hash = ?
                    """,
                    (hash_,),
                )
            )
        return cls.from_dict(
            dict(
                task=loadjson(data[0]),
                host=loadjson(data[1]),
                command=loadjson(data[2]),
                stdout=data[3],
                stderr=data[4],
                returncode=data[5],
                start=data[6],
                end=data[7],
                retry=data[8],
            )
        )

    @classmethod
    def from_dict(cls, dict_: t.Dict[str, t.Any]) -> Task:
        """Overloading from_dict()."""

        return cls(
            **dict(
                dict_,
                command=tuple(dict_["command"]),
                task=Task.from_dict(dict_["task"]),
                host=Host.from_dict(dict_["host"]),
            )
        )

    def to_dict(self) -> t.Dict[str, t.Any]:
        """Overloading to_dict()."""

        return dict(vars(self), task=self.task.to_dict(), host=self.host.to_dict(),)

    def ok(self) -> bool:
        """If the result is success."""
        return self.returncode == 0

    def errored(self) -> bool:
        """If the result is failure."""
        return self.returncode != 0

    def save(self) -> TaskResult:
        """Save the result dump."""

        with ViperDB() as conn:
            conn.execute(
                """
                INSERT INTO task_results (
                    hash, task, host, command, stdout, stderr, returncode, start, end, retry
                ) VALUES (
                    ?, ?, ?, ?, ?, ?, ?, ?, ?, ?
                )
                """,
                (
                    self.hash(),
                    self.task.to_json(),
                    self.host.to_json(),
                    dumpjson(self.command),
                    self.stdout,
                    self.stderr,
                    self.returncode,
                    self.start,
                    self.end,
                    self.retry,
                ),
            )

        return self


@dataclass(frozen=True)
class TaskResults(Items):
    _item_factory: t.Type[TaskResult] = field(init=False, default=TaskResult)

    @classmethod
    def by_host(cls, host: Host) -> TaskResults:
        with ViperDB(ViperDB.url) as conn:
            rows = conn.execute(
                f"""
                SELECT hash FROM task_results
                WHERE JSON_EXTRACT(host, '$.ip') = ?
                ORDER BY start DESC
                """,
                (host.ip,),
            )
            results = [cls._item_factory.from_hash(r[0]) for r in rows]

        return cls.from_items(*results)

    @classmethod
    def by_task(cls, task: Task) -> TaskResults:
        with ViperDB(ViperDB.url) as conn:
            rows = conn.execute(
                f"""
                SELECT hash FROM task_results
                WHERE JSON_EXTRACT(task, '$.name') = ?
                ORDER BY start DESC
                """,
                (task.name,),
            )
            results = [cls._item_factory.from_hash(r[0]) for r in rows]

        return cls.from_items(*results)

    @classmethod
    def by_task_runner(cls, runner: TaskRunner) -> TaskResults:
        with ViperDB(ViperDB.url) as conn:
            rows = conn.execute(
                f"""
                SELECT hash FROM task_results
                WHERE JSON_EXTRACT(host, '$.ip') = ?
                    AND JSON_EXTRACT(task, '$.name') = ?
                ORDER BY start DESC
                """,
                (runner.host.ip, runner.task.name,),
            )
            results = [cls._item_factory.from_hash(r[0]) for r in rows]

        return cls.from_items(*results)

    def hosts(self) -> Hosts:
        """Get the list of hosts from the results."""
        return Hosts.from_items(*(r.host for r in self._all))



"""Viper CLI library."""

import sys
import traceback
import typing as t
from argparse import ArgumentParser
from pydoc import locate

from viper import Hosts, Task, TaskResults, TaskRunners, __version__
from viper.const import Config
from viper.db import ViperDB


def func(objpath):
    """Resolved python function from given string."""

    func = locate(objpath)
    if not func:
        raise ValueError(f"could not resolve {repr(objpath)}.")

    if not callable(func):
        raise ValueError(f"{repr(objpath)} is not a valid function.")

    return func


class SubParser:
    """Base class for the subcommand parsers."""

    subcommand: t.Optional[str] = None

    aliases: t.Sequence[str] = ()

    @classmethod
    def attach_to(cls, subparsers):
        subparser = subparsers.add_parser(cls.subcommand, help=cls.__doc__.lower())
        if cls.aliases:
            for alias in cls.aliases:
                cls(
                    subparsers.add_parser(
                        alias, help=f"alias of {repr(cls.subcommand)}"
                    )
                )
        return cls(subparser)

    def __init__(self, subparser):
        self.add_arguments(subparser)
        subparser.set_defaults(_handler=self)
        subparser.add_argument(
            "--debug",
            action="store_true",
            help="show traceback information when an exception is raised",
        )

    def add_arguments(self, parser):
        raise NotImplementedError()

    def __call__(self, args):
        raise NotImplementedError()


class InitCommand(SubParser):
    """initialize the current workspace"""

    subcommand = "init"

    def add_arguments(self, parser):
        parser.add_argument(
            "-f",
            "--force",
            action="store_true",
            help="remove or overwrite existing data",
        )

    def __call__(self, args) -> int:
        ViperDB.init(ViperDB.url, force=args.force)
        return 0


class TaskFromObjCommand(SubParser):
    """get the task from a Python object location"""

    subcommand = "task:from-obj"

    def add_arguments(self, parser):
        parser.add_argument(
            "obj", type=Task.from_obj, help=Task.from_obj.__doc__.lower()
        )
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args) -> int:
        print(args.obj.to_json(indent=args.indent))
        return 0


class TaskResultsCommand(SubParser):
    """get the past task results of given task"""

    subcommand = "task:results"

    def add_arguments(self, parser):
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args) -> int:
        print(Task.from_json(input()).results().to_json(indent=args.indent))
        return 0


class HostsFromObjCommand(SubParser):
    """get a group of hosts from a Python object location"""

    subcommand = "hosts:from-obj"

    def add_arguments(self, parser):
        parser.add_argument(
            "obj", type=Hosts.from_obj, help=Hosts.from_obj.__doc__.lower()
        )
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args) -> int:
        print(args.obj.to_json(indent=args.indent))
        return 0


class HostsFromFileCommand(SubParser):
    """get a group of hosts from a file"""

    subcommand = "hosts:from-file"

    def add_arguments(self, parser):
        parser.add_argument("filepath")
        parser.add_argument(
            "--loader",
            type=func,
            help="function that resolves a file object to a viper.Hosts object",
        )
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args) -> int:
        print(
            Hosts.from_file(args.filepath, loader=args.loader).to_json(
                indent=args.indent
            )
        )
        return 0


class HostsTaskCommand(SubParser):
    """assign a task to each host"""

    subcommand = "hosts:task"

    def add_arguments(self, parser):
        parser.add_argument(
            "task", type=Task.from_obj, help=Task.from_obj.__doc__.lower()
        )
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args) -> int:

        print(Hosts.from_json(input()).task(args.task).to_json(indent=args.indent))
        return 0


class HostsRunTaskCommand(SubParser):
    """assign a task to each host and run"""

    subcommand = "hosts:run-task"

    def add_arguments(self, parser):
        parser.add_argument(
            "task", type=Task.from_obj, help=Task.from_obj.__doc__.lower()
        )
        parser.add_argument("--max-workers", type=int, default=Config.max_workers.value)
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args) -> int:

        print(
            Hosts.from_json(input())
            .run_task(args.task, max_workers=args.max_workers)
            .to_json(indent=args.indent)
        )
        return 0


class HostsRunTaskThenPipeCommand(SubParser):
    """run the given task on the hosts and then pipe the result to the given function"""

    subcommand = "hosts:run-task-then-pipe"

    def add_arguments(self, parser):
        parser.add_argument(
            "task", type=Task.from_obj, help=Task.from_obj.__doc__.lower()
        )
        parser.add_argument("handler", type=func, help="the result handler function")
        parser.add_argument("--max-workers", type=int, default=Config.max_workers.value)

    def __call__(self, args) -> int:
        Hosts.from_json(input()).run_task_then_pipe(
            args.task, args.handler, max_workers=args.max_workers
        )
        return 0


class HostsFilterCommand(SubParser):
    """filter hosts by a given function"""

    subcommand = "hosts:filter"

    def add_arguments(self, parser):
        parser.add_argument("filter", type=func)
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args):
        print(Hosts.from_json(input()).filter(args.filter).to_json(indent=args.indent))
        return 0


class HostsCountCommand(SubParser):
    """count the number of hosts"""

    subcommand = "hosts:count"

    def add_arguments(self, parser):
        pass

    def __call__(self, args):
        print(Hosts.from_json(input()).count())
        return 0


class HostsSortCommand(SubParser):
    """sort the hosts"""

    subcommand = "hosts:sort"

    def add_arguments(self, parser):
        parser.add_argument("--key", type=func)
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args):
        print(Hosts.from_json(input()).sort(key=args.key).to_json(indent=args.indent))
        return 0


class HostsPipeCommand(SubParser):
    """pipe the hosts to the given function"""

    subcommand = "hosts:pipe"

    def add_arguments(self, parser):
        parser.add_argument("func", type=func)

    def __call__(self, args) -> int:
        Hosts.from_json(input()).pipe(args.func)
        return 0


class HostsTaskResultsCommand(SubParser):
    """get the past task results of the hosts"""

    subcommand = "hosts:task-results"

    def add_arguments(self, parser):
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args):
        print(Hosts.from_json(input()).task_results().to_json(indent=args.indent))
        return 0


class TaskRunnersFilterCommand(SubParser):
    """filter task runners by a given function"""

    subcommand = "task-runners:filter"

    def add_arguments(self, parser):
        parser.add_argument("filter", type=func)
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args):
        print(
            TaskRunners.from_json(input())
            .filter(args.filter)
            .to_json(indent=args.indent)
        )
        return 0


class TaskRunnersCountCommand(SubParser):
    """count the number of task runners"""

    subcommand = "task-runners:count"

    def add_arguments(self, parser):
        pass

    def __call__(self, args):
        print(TaskRunners.from_json(input()).count())
        return 0


class TaskRunnersSortCommand(SubParser):
    """sort the task runners"""

    subcommand = "task-runners:sort"

    def add_arguments(self, parser):
        parser.add_argument("--key", type=func)
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args):
        print(
            TaskRunners.from_json(input())
            .sort(key=args.key)
            .to_json(indent=args.indent)
        )
        return 0


class TaskRunnersPipeCommand(SubParser):
    """pipe the task runners to the given function"""

    subcommand = "task-runners:pipe"

    def add_arguments(self, parser):
        parser.add_argument("func", type=func)

    def __call__(self, args) -> int:
        TaskRunners.from_json(input()).pipe(args.func)
        return 0


class TaskRunnersRunCommand(SubParser):
    """run the assigned tasks"""

    subcommand = "task-runners:run"

    def add_arguments(self, parser):
        parser.add_argument("--max-workers", type=int, default=Config.max_workers.value)
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args):
        print(
            TaskRunners.from_json(input())
            .run(max_workers=args.max_workers)
            .to_json(indent=args.indent)
        )
        return 0


class TaskRunnersHostsCommand(SubParser):
    """get the hohsts from the task runners"""

    subcommand = "task-runners:hosts"

    def add_arguments(self, parser):
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args):
        print(TaskRunners.from_json(input()).hosts().to_json(indent=args.indent))
        return 0


class TaskResultsFilterCommand(SubParser):
    """filter task results by a given function"""

    subcommand = "task-results:filter"

    def add_arguments(self, parser):
        parser.add_argument("filter", type=func)
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args):
        print(
            TaskResults.from_json(input())
            .filter(args.filter)
            .to_json(indent=args.indent)
        )
        return 0


class TaskResultsCountCommand(SubParser):
    """count the number of task results"""

    subcommand = "task-results:count"

    def add_arguments(self, parser):
        pass

    def __call__(self, args):
        print(TaskResults.from_json(input()).count())
        return 0


class TaskResultsSortCommand(SubParser):
    """sort the task results"""

    subcommand = "task-results:sort"

    def add_arguments(self, parser):
        parser.add_argument("--key", type=func)
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args):
        print(
            TaskResults.from_json(input())
            .sort(key=args.key)
            .to_json(indent=args.indent)
        )
        return 0


class TaskResultsPipeCommand(SubParser):
    """pipe the task results to the given function"""

    subcommand = "task-results:pipe"

    def add_arguments(self, parser):
        parser.add_argument("func", type=func)

    def __call__(self, args) -> int:
        TaskResults.from_json(input()).pipe(args.func)
        return 0


class TaskResultsHostsCommand(SubParser):
    """get the hosts from the task results"""

    subcommand = "task-results:hosts"

    def add_arguments(self, parser):
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args):
        print(TaskResults.from_json(input()).hosts().to_json(indent=args.indent))
        return 0


class TaskResultsByTaskCommand(SubParser):
    """get the past task results of given task"""

    subcommand = "task-results:by-task"

    def add_arguments(self, parser):
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args):
        print(TaskResults.by_task(Task.from_json(input())).to_json(indent=args.indent))
        return 0


def run() -> int:
    parser = ArgumentParser("viper", description=f"Viper CLI {__version__}")
    parser.add_argument("--version", action="version", version=__version__)
    parser.add_argument(
        "--debug",
        action="store_true",
        help="show traceback information when an exception is raised",
    )
    subparsers = parser.add_subparsers()

    # Init command
    InitCommand.attach_to(subparsers)

    # Task commands
    TaskFromObjCommand.attach_to(subparsers)
    TaskResultsCommand.attach_to(subparsers)

    # Hosts commands
    HostsFromFileCommand.attach_to(subparsers)
    HostsFromObjCommand.attach_to(subparsers)

    HostsFilterCommand.attach_to(subparsers)
    HostsCountCommand.attach_to(subparsers)
    HostsSortCommand.attach_to(subparsers)
    HostsPipeCommand.attach_to(subparsers)

    HostsTaskCommand.attach_to(subparsers)
    HostsRunTaskCommand.attach_to(subparsers)
    HostsRunTaskThenPipeCommand.attach_to(subparsers)
    HostsTaskResultsCommand.attach_to(subparsers)

    # Task runners commands
    TaskRunnersFilterCommand.attach_to(subparsers)
    TaskRunnersCountCommand.attach_to(subparsers)
    TaskRunnersSortCommand.attach_to(subparsers)
    TaskRunnersPipeCommand.attach_to(subparsers)

    TaskRunnersRunCommand.attach_to(subparsers)
    TaskRunnersHostsCommand.attach_to(subparsers)

    # Task results commands
    TaskResultsFilterCommand.attach_to(subparsers)
    TaskResultsCountCommand.attach_to(subparsers)
    TaskResultsSortCommand.attach_to(subparsers)
    TaskResultsPipeCommand.attach_to(subparsers)

    TaskResultsHostsCommand.attach_to(subparsers)
    TaskResultsByTaskCommand.attach_to(subparsers)

    args = parser.parse_args()

    if not hasattr(args, "_handler"):
        parser.print_usage()
        return 2

    try:
        return args._handler(args)
    except Exception as e:
        print(f"error: {e}", file=sys.stderr)
        if args.debug:
            print(traceback.format_exc(), file=sys.stderr)
        return 1



"""## Viper CLI Examples.

### Initialize current workspace (creates a `viperdb.sqlite3` file)

    viper init

    # or remove or overwrite existing data with "-f" / "--force"

    viper init -f


### Load hosts from file

    viper hosts:from-file tests/data/hosts.csv --indent 4

    # or with a custom loader

    viper hosts:from-file tests/data/hosts.json --loader viper.demo.loaders.json --indent 4


### Load hosts from Python object

    viper hosts:from-obj viper.demo.hosts.group1 --indent 4


### Let's save the hosts

    viper hosts:from-obj viper.demo.hosts.group1 > /tmp/hosts.json


### Filter hosts

    cat /tmp/hosts.json | viper hosts:filter viper.demo.filters.ip_starts_with_2 --indent 4


### Assign tasks to the given hosts

    cat /tmp/hosts.json | viper hosts:task viper.demo.tasks.ping --indent 4


### Run the assigned tasks

    cat /tmp/hosts.json | viper hosts:task viper.demo.tasks.ping | viper task-runners:run --indent 4

    # or use a shortcut

    cat /tmp/hosts.json | viper hosts:run-task viper.demo.tasks.ping --indent 4


### Run tasks in parallel using multiple workers

    cat /tmp/hosts.json | viper hosts:run-task viper.demo.tasks.ping --max-workers 50 --indent 4


### Get the past task results of the hosts from DB

    cat /tmp/hosts.json | viper hosts:task-results --indent 4 --debug


### Or get the past task results by task

    viper task:from-obj viper.demo.tasks.ping | viper task-results:by-task -i 4

    # Or

    viper task:from-obj viper.demo.tasks.ping | viper task:results -i 4


### Let's save the result

    viper task:from-obj viper.demo.tasks.ping | viper task:results > /tmp/results.json


### Now filter the results by their status

    # success
    cat /tmp/results.json | viper task-results:filter viper.demo.filters.result_ok -i 4

    # failed
    cat /tmp/results.json | viper task-results:filter viper.demo.filters.result_errored -i 4


### Pipe the results to a custom handler

    cat /tmp/results.json | viper task-results:pipe viper.demo.handlers.print_status


### Let's do that again in one go
    viper hosts:from-obj viper.demo.hosts.group1 | viper hosts:run-task-then-pipe viper.demo.tasks.ping viper.demo.handlers.print_status
"""



from viper import Host, TaskResult
from viper.demo import __doc__


def ip_starts_with_2(host: Host) -> bool:
    """Filter hosts IPs that start with 2."""
    return host.ip.startswith("2")


def result_ok(result: TaskResult) -> bool:
    """Filter task results by OK status."""
    return result.ok()


def result_errored(result: TaskResult) -> bool:
    """Filter task results by error status."""
    return result.errored()
